(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (factory((global.redom = {})));
}(this, (function (exports) { 'use strict';

  var HASH = '#'.charCodeAt(0);
  var DOT = '.'.charCodeAt(0);

  var TAG_NAME = 0;
  var ID = 1;
  var CLASS_NAME = 2;

  var parseQuery = function (query) {
    var tag = null;
    var id = null;
    var className = null;
    var mode = TAG_NAME;
    var offset = 0;

    for (var i = 0; i <= query.length; i++) {
      var char = query.charCodeAt(i);
      var isHash = char === HASH;
      var isDot = char === DOT;
      var isEnd = !char;

      if (isHash || isDot || isEnd) {
        if (mode === TAG_NAME) {
          if (i === 0) {
            tag = 'div';
          } else {
            tag = query.substring(offset, i);
          }
        } else if (mode === ID) {
          id = query.substring(offset, i);
        } else {
          if (className) {
            className += ' ' + query.substring(offset, i);
          } else {
            className = query.substring(offset, i);
          }
        }

        if (isHash) {
          mode = ID;
        } else if (isDot) {
          mode = CLASS_NAME;
        }

        offset = i + 1;
      }
    }

    return { tag: tag, id: id, className: className };
  };

  var createElement = function (query, ns) {
    var ref = parseQuery(query);
    var tag = ref.tag;
    var id = ref.id;
    var className = ref.className;
    var element = ns ? document.createElementNS(ns, tag) : document.createElement(tag);

    if (id) {
      element.id = id;
    }

    if (className) {
      if (ns) {
        element.setAttribute('class', className);
      } else {
        element.className = className;
      }
    }

    return element;
  };

  var unmount = function (parent, child) {
    var parentEl = getEl(parent);
    var childEl = getEl(child);

    if (child === childEl && childEl.__redom_view) {
      // try to look up the view if not provided
      child = childEl.__redom_view;
    }

    if (childEl.parentNode) {
      doUnmount(child, childEl, parentEl);

      parentEl.removeChild(childEl);
    }

    return child;
  };

  var doUnmount = function (child, childEl, parentEl) {
    var hooks = childEl.__redom_lifecycle;

    if (hooksAreEmpty(hooks)) {
      childEl.__redom_mounted = false;
      return;
    }

    var traverse = parentEl;

    if (childEl.__redom_mounted) {
      trigger(childEl, 'onunmount');
    }

    while (traverse) {
      var parentHooks = traverse.__redom_lifecycle || {};

      for (var hook in hooks) {
        if (parentHooks[hook]) {
          parentHooks[hook] -= hooks[hook];
        }
      }

      if (hooksAreEmpty(parentHooks)) {
        traverse.__redom_lifecycle = null;
      }

      traverse = traverse.parentNode;
    }
  };

  var hooksAreEmpty = function (hooks) {
    return !hooks || !Object.keys(hooks).filter(function (hook) { return hooks[hook]; }).length;
  };

  var hookNames = ['onmount', 'onunmount'];
  var shadowRootAvailable = typeof window !== 'undefined' && 'ShadowRoot' in window;

  var mount = function (parent, child, before) {
    var parentEl = getEl(parent);
    var childEl = getEl(child);

    if (child === childEl && childEl.__redom_view) {
      // try to look up the view if not provided
      child = childEl.__redom_view;
    }

    if (child !== childEl) {
      childEl.__redom_view = child;
    }

    var wasMounted = childEl.__redom_mounted;
    var oldParent = childEl.parentNode;

    if (wasMounted && (oldParent !== parentEl)) {
      doUnmount(child, childEl, oldParent);
    }

    if (before != null) {
      parentEl.insertBefore(childEl, getEl(before));
    } else {
      parentEl.appendChild(childEl);
    }

    doMount(child, childEl, parentEl, oldParent);

    return child;
  };

  var doMount = function (child, childEl, parentEl, oldParent) {
    var hooks = childEl.__redom_lifecycle || (childEl.__redom_lifecycle = {});
    var remount = (parentEl === oldParent);
    var hooksFound = false;

    for (var i = 0; i < hookNames.length; i++) {
      var hookName = hookNames[i];

      if (!remount && (child !== childEl) && (hookName in child)) {
        hooks[hookName] = (hooks[hookName] || 0) + 1;
      }
      if (hooks[hookName]) {
        hooksFound = true;
      }
    }

    if (!hooksFound) {
      childEl.__redom_mounted = true;
      return;
    }

    var traverse = parentEl;
    var triggered = false;

    if (remount || (!triggered && (traverse && traverse.__redom_mounted))) {
      trigger(childEl, remount ? 'onremount' : 'onmount');
      triggered = true;
    }

    if (remount) {
      return;
    }

    while (traverse) {
      var parent = traverse.parentNode;
      var parentHooks = traverse.__redom_lifecycle || (traverse.__redom_lifecycle = {});

      for (var hook in hooks) {
        parentHooks[hook] = (parentHooks[hook] || 0) + hooks[hook];
      }

      if (!triggered && (traverse === document || (shadowRootAvailable && (traverse instanceof window.ShadowRoot)) || (parent && parent.__redom_mounted))) {
        trigger(traverse, remount ? 'onremount' : 'onmount');
        triggered = true;
      }

      traverse = parent;
    }
  };

  var trigger = function (el, eventName) {
    if (eventName === 'onmount') {
      el.__redom_mounted = true;
    } else if (eventName === 'onunmount') {
      el.__redom_mounted = false;
    }

    var hooks = el.__redom_lifecycle;

    if (!hooks) {
      return;
    }

    var view = el.__redom_view;
    var hookCount = 0;

    view && view[eventName] && view[eventName]();

    for (var hook in hooks) {
      if (hook) {
        hookCount++;
      }
    }

    if (hookCount) {
      var traverse = el.firstChild;

      while (traverse) {
        var next = traverse.nextSibling;

        trigger(traverse, eventName);

        traverse = next;
      }
    }
  };

  var setStyle = function (view, arg1, arg2) {
    var el = getEl(view);

    if (arg2 !== undefined) {
      el.style[arg1] = arg2;
    } else if (isString(arg1)) {
      el.setAttribute('style', arg1);
    } else {
      for (var key in arg1) {
        setStyle(el, key, arg1[key]);
      }
    }
  };

  /* global SVGElement */

  var xlinkns = 'http://www.w3.org/1999/xlink';

  var setAttr = function (view, arg1, arg2) {
    var el = getEl(view);
    var isSVG = el instanceof SVGElement;

    if (arg2 !== undefined) {
      if (arg1 === 'style') {
        setStyle(el, arg2);
      } else if (isSVG && isFunction(arg2)) {
        el[arg1] = arg2;
      } else if (arg1 === 'dataset') {
        setData(el, arg2);
      } else if (!isSVG && (arg1 in el || isFunction(arg2))) {
        el[arg1] = arg2;
      } else {
        if (isSVG && (arg1 === 'xlink')) {
          setXlink(el, arg2);
          return;
        }
        el.setAttribute(arg1, arg2);
      }
    } else {
      for (var key in arg1) {
        setAttr(el, key, arg1[key]);
      }
    }
  };

  function setXlink (el, obj) {
    for (var key in obj) {
      el.setAttributeNS(xlinkns, key, obj[key]);
    }
  }

  function setData (el, obj) {
    for (var key in obj) {
      el.dataset[key] = obj[key];
    }
  }

  var text = function (str) { return document.createTextNode((str != null) ? str : ''); };

  var parseArguments = function (element, args) {
    for (var i = 0; i < args.length; i++) {
      var arg = args[i];

      if (arg !== 0 && !arg) {
        continue;
      }

      // support middleware
      if (typeof arg === 'function') {
        arg(element);
      } else if (isString(arg) || isNumber(arg)) {
        element.appendChild(text(arg));
      } else if (isNode(getEl(arg))) {
        mount(element, arg);
      } else if (arg.length) {
        parseArguments(element, arg);
      } else if (typeof arg === 'object') {
        setAttr(element, arg);
      }
    }
  };

  var ensureEl = function (parent) { return isString(parent) ? html(parent) : getEl(parent); };
  var getEl = function (parent) { return (parent.nodeType && parent) || (!parent.el && parent) || getEl(parent.el); };

  var isString = function (a) { return typeof a === 'string'; };
  var isNumber = function (a) { return typeof a === 'number'; };
  var isFunction = function (a) { return typeof a === 'function'; };

  var isNode = function (a) { return a && a.nodeType; };

  var htmlCache = {};

  var memoizeHTML = function (query) { return htmlCache[query] || (htmlCache[query] = createElement(query)); };

  var html = function (query) {
    var args = [], len = arguments.length - 1;
    while ( len-- > 0 ) args[ len ] = arguments[ len + 1 ];

    var element;

    if (isString(query)) {
      element = memoizeHTML(query).cloneNode(false);
    } else if (isNode(query)) {
      element = query.cloneNode(false);
    } else {
      throw new Error('At least one argument required');
    }

    parseArguments(element, args);

    return element;
  };

  html.extend = function (query) {
    var args = [], len = arguments.length - 1;
    while ( len-- > 0 ) args[ len ] = arguments[ len + 1 ];

    var clone = memoizeHTML(query);

    return html.bind.apply(html, [ this, clone ].concat( args ));
  };

  var el = html;
  var h = html;

  var setChildren = function (parent) {
    var children = [], len = arguments.length - 1;
    while ( len-- > 0 ) children[ len ] = arguments[ len + 1 ];

    var parentEl = getEl(parent);
    var current = traverse(parent, children, parentEl.firstChild);

    while (current) {
      var next = current.nextSibling;

      unmount(parent, current);

      current = next;
    }
  };

  function traverse (parent, children, _current) {
    var current = _current;

    for (var i = 0; i < children.length; i++) {
      var child = children[i];

      if (!child) {
        continue;
      }

      var childEl = getEl(child);

      if (childEl === current) {
        current = current.nextSibling;
        continue;
      }

      if (isNode(childEl)) {
        mount(parent, child, current);
        continue;
      }

      if (child.length != null) {
        current = traverse(parent, child, current);
      }
    }

    return current;
  }

  var propKey = function (key) { return function (item) { return item[key]; }; };

  var listPool = function (View, key, initData) {
    return new ListPool(View, key, initData);
  };

  var ListPool = function ListPool (View, key, initData) {
    this.View = View;
    this.initData = initData;
    this.oldLookup = {};
    this.lookup = {};
    this.oldViews = [];
    this.views = [];

    if (key != null) {
      this.lookup = {};
      this.key = isFunction(key) ? key : propKey(key);
    }
  };
  ListPool.prototype.update = function update (data, context) {
    var ref = this;
      var View = ref.View;
      var key = ref.key;
      var initData = ref.initData;
    var keySet = key != null;

    var oldLookup = this.lookup;
    var newLookup = {};

    var newViews = new Array(data.length);
    var oldViews = this.views;

    for (var i = 0; i < data.length; i++) {
      var item = data[i];
      var view = (void 0);

      if (keySet) {
        var id = key(item);
        view = oldLookup[id] || new View(initData, item, i, data);
        newLookup[id] = view;
        view.__redom_id = id;
      } else {
        view = oldViews[i] || new View(initData, item, i, data);
      }
      view.update && view.update(item, i, data, context);

      var el = getEl(view.el);
      el.__redom_view = view;
      newViews[i] = view;
    }

    this.oldViews = oldViews;
    this.views = newViews;

    this.oldLookup = oldLookup;
    this.lookup = newLookup;
  };

  var list = function (parent, View, key, initData) {
    return new List(parent, View, key, initData);
  };

  var List = function List (parent, View, key, initData) {
    this.__redom_list = true;
    this.View = View;
    this.initData = initData;
    this.views = [];
    this.pool = new ListPool(View, key, initData);
    this.el = ensureEl(parent);
    this.keySet = key != null;
  };
  List.prototype.update = function update (data, context) {
      var this$1 = this;
      if ( data === void 0 ) data = [];

    var ref = this;
      var keySet = ref.keySet;
    var oldViews = this.views;
    var oldLookup = keySet && this.lookup;

    this.pool.update(data, context);
    var ref$1 = this.pool;
      var views = ref$1.views;
      var lookup = ref$1.lookup;

    if (keySet) {
      for (var i = 0; i < oldViews.length; i++) {
        var id = oldViews[i].__redom_id;
        if (!(id in lookup)) {
          unmount(this$1, oldLookup[id]);
        }
      }
    }

    setChildren(this, views);

    if (keySet) {
      this.lookup = lookup;
    }
    this.views = views;
  };

  List.extend = function (parent, View, key, initData) {
    return List.bind(List, parent, View, key, initData);
  };

  list.extend = List.extend;

  /* global Node */

  var place = function (View, initData) {
    return new Place(View, initData);
  };

  var Place = function Place (View, initData) {
    this.el = text('');
    this.visible = false;
    this.view = null;
    this._placeholder = this.el;
    if (View instanceof Node) {
      this._el = View;
    } else {
      this._View = View;
    }
    this._initData = initData;
  };
  Place.prototype.update = function update (visible, data) {
    var placeholder = this._placeholder;
    var parentNode = this.el.parentNode;

    if (visible) {
      if (!this.visible) {
        if (this._el) {
          mount(parentNode, this._el, placeholder);
          unmount(parentNode, placeholder);
          this.el = this._el;
          this.visible = visible;
          return;
        }
        var View = this._View;
        var view = new View(this._initData);

        this.el = getEl(view);
        this.view = view;

        mount(parentNode, view, placeholder);
        unmount(parentNode, placeholder);
      }
      this.view && this.view.update && this.view.update(data);
    } else {
      if (this.visible) {
        if (this._el) {
          mount(parentNode, placeholder, this._el);
          unmount(parentNode, this._el);
          this.el = placeholder;
          this.visible = visible;
          return;
        }
        mount(parentNode, placeholder, this.view);
        unmount(parentNode, this.view);

        this.el = placeholder;
        this.view = null;
      }
    }
    this.visible = visible;
  };

  var router = function (parent, Views, initData) {
    return new Router(parent, Views, initData);
  };

  var Router = function Router (parent, Views, initData) {
    this.el = ensureEl(parent);
    this.Views = Views;
    this.initData = initData;
  };
  Router.prototype.update = function update (route, data) {
    if (route !== this.route) {
      var Views = this.Views;
      var View = Views[route];

      this.route = route;
      this.view = View && new View(this.initData, data);

      setChildren(this.el, [ this.view ]);
    }
    this.view && this.view.update && this.view.update(data, route);
  };

  var ns = 'http://www.w3.org/2000/svg';

  var svgCache = {};

  var memoizeSVG = function (query) { return svgCache[query] || (svgCache[query] = createElement(query, ns)); };

  var svg = function (query) {
    var args = [], len = arguments.length - 1;
    while ( len-- > 0 ) args[ len ] = arguments[ len + 1 ];

    var element;

    if (isString(query)) {
      element = memoizeSVG(query).cloneNode(false);
    } else if (isNode(query)) {
      element = query.cloneNode(false);
    } else {
      throw new Error('At least one argument required');
    }

    parseArguments(element, args);

    return element;
  };

  svg.extend = function (query) {
    var clone = memoizeSVG(query);

    return svg.bind(this, clone);
  };

  svg.ns = ns;

  var s = svg;

  exports.el = el;
  exports.h = h;
  exports.html = html;
  exports.list = list;
  exports.List = List;
  exports.listPool = listPool;
  exports.ListPool = ListPool;
  exports.mount = mount;
  exports.unmount = unmount;
  exports.place = place;
  exports.Place = Place;
  exports.router = router;
  exports.Router = Router;
  exports.setAttr = setAttr;
  exports.setStyle = setStyle;
  exports.setChildren = setChildren;
  exports.s = s;
  exports.svg = svg;
  exports.text = text;

  Object.defineProperty(exports, '__esModule', { value: true });

})));
